/*
 * This file is part of xine and vlc , unix video players.
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * It is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * Done by  Maupatz 2001
 */
#ifdef HAVE_CONFIG_H
#include  "config.h"
#endif

#include "defs.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dlfcn.h>
#include <inttypes.h>
#include <unistd.h> 
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h> 

#if defined (WIN32)
#include <windows.h>
#include <io.h>
#include <winioctl.h>
#include "langid.h"
#else
#include <sys/ioctl.h>
#include <sys/poll.h>
#include <sys/stat.h>
#endif

#if defined (HAVE_SYS_CDIO_H)
# include <sys/cdio.h>
#endif

#if defined (HAVE_LINUX_CDROM_H)
# include <linux/cdrom.h>
#endif

#if !defined (HAVE_SYS_CDIO_H) && !defined (HAVE_LINUX_CDROM_H) && !defined (WIN32)
# error "You need the DVD/CDROM headers in your system"
#endif

#include "dvd_ifo.h"
#include "dvd_css.h"
#include "dvd_ioctl.h"
#include "input_dvd.h"
#include "dvd_udf.h"
#include "read_cache.h"
#include "parse_dsi.h"
#include "osfileio.h"

#include <xine/xine_internal.h>
#include <xine/xineutils.h>
#include <xine/buffer.h>
#include <xine/input_plugin.h>
#include <xine/video_out.h>
#include <xine/events.h>
#include <xine/spu_decoder_api.h>

#if defined (__sun)
#define RDVD    "/vol/dev/aliases/cdrom0"
#define DVD     RDVD
#elif defined (WIN32)
#define RDVD "E:"
#define DVD RDVD
#else
#define DVD     "/dev/dvd"
#define RDVD    "/dev/rdvd"
#endif

#define DVD_VIDEO_LB_LEN  DVD_LB_SIZE

/*
 * To avoid freeze
 */

#define CHAPTER_SIZE_MIN 16

/*
 * global Variables:
 */

typedef struct {

  input_plugin_t    input_plugin;

  char             mrl[256];
  config_values_t  *config;
  read_cache_t     *read_cache;

  const char       *device;
  const char       *raw_device;

  int               dvd_fd;
  int               raw_fd;

  /*
   * udf dir function
   */

#define MAX_DIR_ENTRIES 1024

  mrl_t            *mrls[MAX_DIR_ENTRIES];
  char             *filelist[MAX_DIR_ENTRIES];

  int NENTRIES;

  char title_display[256];
  int need_update_title;

  thread_dvd_data_t  *p_dvd;

  dsi_t dsi;
  int dsi_valid;
  int dsi_pos;
  int angle_change_request;
  int next_angle_change;
  int angle_change_enabled;
  int next_angle_ilvu_start;
  int next_ilvu_start;
  int angle_available;

  int branching; /* keep angle if branching */
  int xine_is_seeking;

  int is_movie;

  struct {
          int title;
          int start_chapter;
          int end_chapter;
          int angle;
  } movie;

  /* global variables for a chapter */
#define MAX_ANGLE 9
  int current_size[MAX_ANGLE + 1];
  int current_pos;
  int current_chapter;
  int current_title;
  int current_angle;
  int current_sector_end; /* hack for calculate the real length */
  int free_ifo;

  int first_play; /* To select the language at startup */
  uint16_t LANG;  /* language at startup */

  xine_t           *xine;

} dvd_input_plugin_t;

/* ***************************************************************** */
/*                        Private functions                          */
/* ***************************************************************** */

/*
 * Callbacks for configuration changes.
 */

static void device_change_cb(void *data, cfg_entry_t *cfg) {
  dvd_input_plugin_t *this = (dvd_input_plugin_t *) data;

  this->device = strdup(cfg->str_value);
}

static void rawdevice_change_cb(void *data, cfg_entry_t *cfg) {
  dvd_input_plugin_t *this = (dvd_input_plugin_t *) data;

  this->raw_device = strdup(cfg->str_value);
}

static int openDrive (dvd_input_plugin_t *this) {

  this->dvd_fd = dvd_os_open(this->device);

  if (this->dvd_fd == -1) {
    printf ("input_dmd: unable to open dvd drive (%s): %s\n",this->device ,
	    strerror(errno));
    return -1;
  }
#ifdef WIN32
  this->raw_fd=this->dvd_fd;
#else
  this->raw_fd = dvd_os_open(this->raw_device);
  if (this->raw_fd ==-1) {
    this->raw_fd = this->dvd_fd;
  } else {
    printf ("input_dmd: dmd will use the raw device : %s\n",this->raw_device);
  }
#endif

  read_cache_set_fd (this->read_cache, this->raw_fd);

  return this->raw_fd;
}

static void closeDrive (dvd_input_plugin_t *this) {

 if (this->dvd_fd ==-1) return;

 dvd_os_close(this->dvd_fd);

 if (this->raw_fd != this->dvd_fd)  dvd_os_close(this->raw_fd);

 this->dvd_fd = this->raw_fd = -1;
}

#ifdef __sun
/*
 * Check the environment, if we're running under sun's
 * vold/rmmount control.
 */
static void
check_solaris_vold_device(dvd_input_plugin_t *this)
{
  char *volume_device;
  char *volume_name;
  char *volume_action;
  char *device;
  struct stat stb;

  if ((volume_device = getenv("VOLUME_DEVICE")) != NULL &&
      (volume_name   = getenv("VOLUME_NAME"))   != NULL &&
      (volume_action = getenv("VOLUME_ACTION")) != NULL &&
      strcmp(volume_action, "insert") == 0) {

    device = malloc(strlen(volume_device) + strlen(volume_name) + 2);
    if (device == NULL)
      return;
    sprintf(device, "%s/%s", volume_device, volume_name);
    if (stat(device, &stb) != 0 || !S_ISCHR(stb.st_mode)) {
      free(device);
      return;
    }
    this->device = this->raw_device = device;
  }
}
#endif

/* ***************************************************************** */
/*                         END OF PRIVATES                           */
/* ***************************************************************** */

/**
 * Special fuction to show title
 */

void update_title_display(dvd_input_plugin_t *this) {
  xine_ui_event_t uevent;

  if(!this || !(this->xine))
   return;

  /* Set title/chapter display */
  uevent.event.type = XINE_EVENT_UI_SET_TITLE;
  uevent.data = this->title_display;

    /* set title */

  if(this->p_dvd->i_angle_nb == 1 )
    {
  /* Need to set the name of the movie instead */
      if (this->is_movie)
      // snprintf(this->title_display, 255,"Movie");
         snprintf(this->title_display, 255,UDFVolumeDesc());
      else
      snprintf(this->title_display, 255,
	       "Title %i - Chapter %i", this->current_title,this->current_chapter);
    }
     else
    {
     if (this->is_movie)
     /*
     snprintf(this->title_display, 255,"Movie - Angle %d/%d",
              this->current_angle,
              this->p_dvd->i_angle_nb );
     */
      snprintf(this->title_display, 255,"%s - Angle %d/%d",
              UDFVolumeDesc(),
              this->current_angle,
              this->p_dvd->i_angle_nb );
     else
     snprintf(this->title_display, 255,
	       "Title %i - Chapter %i - Angle %d/%d",
               this->current_title,
               this->current_chapter,
               this->current_angle,
               this->p_dvd->i_angle_nb);
     if (this->angle_available) strncat(this->title_display," - Have angles",255);
     else strncat(this->title_display," - No angle",255);
    }

  xine_send_event(this->xine, &uevent.event);
}

/**
 * Special fuction to show title
 */

#define title \
    this->p_dvd->p_ifo->vts.title_unit.p_title[this->p_dvd->i_title_id-1].title

#define cell  this->p_dvd->p_ifo->vts.cell_inf

/*****************************************************************************
 * DVDChooseAngle: select the cell corresponding to the selected angle
 *****************************************************************************/
static int DVDChooseAngle( dvd_input_plugin_t *this)
{

    /* basic handling of angles */

    switch( ( ( title.p_cell_play[this->p_dvd->i_prg_cell].i_category & 0xf000 )
                    >> 12 ) )
    {
        /* we enter a muli-angle section */
        case 0x5:
            this->p_dvd->i_prg_cell += this->p_dvd->i_angle - 1;
//            printf( 1, "dmd info: choosing angle %d", this->p_dvd->i_angle );
            break;
        /* we exit a multi-angle section */
        case 0x9:
        case 0xd:
            this->p_dvd->i_prg_cell += this->p_dvd->i_angle_nb - this->p_dvd->i_angle;
            break;
    }

    return 0;
}

/*****************************************************************************
 * DVDFindCell: adjust the title cell index with the program cell
 *****************************************************************************/
static int DVDFindCell(dvd_input_plugin_t *this  )
{
    int                 i_cell;
    int                 i_index;

    i_cell = this->p_dvd->i_cell;
    i_index = this->p_dvd->i_prg_cell;

    if( i_cell >= cell.i_cell_nb )
    {
        return -1;
    }

    while( ( ( title.p_cell_pos[i_index].i_vob_id !=
                   cell.p_cell_map[i_cell].i_vob_id ) ||
      ( title.p_cell_pos[i_index].i_cell_id !=
                   cell.p_cell_map[i_cell].i_cell_id ) ) &&
           ( i_cell < cell.i_cell_nb - 1 ) )
    {
        i_cell++;
    }

 /*
printf( "FindCell: i_cell %d i_index %d found %d nb %d",
                    this->p_dvd->i_cell,
                    this->p_dvd->i_prg_cell,
                    i_cell,
                    cell.i_cell_nb );

*/
    this->p_dvd->i_cell = i_cell;

    return 0;
}

#undef cell

/*****************************************************************************
 * DVDFindSector: find cell index in adress map from index in
 * information table program map and give corresponding sectors.
 *****************************************************************************/
static  int DVDFindSector( dvd_input_plugin_t *this )
{

    if( this->p_dvd->i_sector > (int) title.p_cell_play[this->p_dvd->i_prg_cell].i_end_sector )
    {
       this->p_dvd->i_prg_cell++;

       if( DVDChooseAngle(this) < 0 )
        {
            return -1;
        }

    }

    if( DVDFindCell(this) < 0 )
    {
//        printf( "dmd error: can't find sector\n" );
        return -1;
    }

    /* Find start and end sectors of new cell */

    this->p_dvd->i_sector = MAX(
         this->p_dvd->p_ifo->vts.cell_inf.p_cell_map[this->p_dvd->i_cell].i_start_sector,
         title.p_cell_play[this->p_dvd->i_prg_cell].i_start_sector );
    this->p_dvd->i_end_sector = MIN(
         this->p_dvd->p_ifo->vts.cell_inf.p_cell_map[this->p_dvd->i_cell].i_end_sector,
         title.p_cell_play[this->p_dvd->i_prg_cell].i_end_sector );

    /*
    printf( "cell: %d sector1: 0x%x end1: 0x%x\n"
                   "index: %d sector2: 0x%x end2: 0x%x\n"
                   "category: 0x%x ilvu end: 0x%x vobu start 0x%x",
        this->p_dvd->i_cell,
        this->p_dvd->p_ifo->vts.cell_inf.p_cell_map[this->p_dvd->i_cell].i_start_sector,
        this->p_dvd->p_ifo->vts.cell_inf.p_cell_map[this->p_dvd->i_cell].i_end_sector,
        this->p_dvd->i_prg_cell,
        title.p_cell_play[this->p_dvd->i_prg_cell].i_start_sector,
        title.p_cell_play[this->p_dvd->i_prg_cell].i_end_sector,
        title.p_cell_play[this->p_dvd->i_prg_cell].i_category,
        title.p_cell_play[this->p_dvd->i_prg_cell].i_first_ilvu_vobu_esector,
        title.p_cell_play[this->p_dvd->i_prg_cell].i_last_vobu_start_sector );
      */

    return 0;
}

 /*****************************************************************************
 * DVDChapterSelect: find the cell corresponding to requested chapter
 *****************************************************************************/
static int DVDChapterSelect( dvd_input_plugin_t *this, int i_chapter )
{

    /* Find cell index in Program chain for current chapter */
    this->p_dvd->i_prg_cell = title.chapter_map.pi_start_cell[i_chapter-1] - 1;
    this->p_dvd->i_cell = 0;
    this->p_dvd->i_sector = 0;

    DVDChooseAngle(this);

    /* Search for cell_index in cell adress_table and initialize
     * start sector */
    if( DVDFindSector(this) < 0 )
    {
  //      printf( "dmd error: can't select chapter \n" );
        return -1;
    }

    /* start is : beginning of vts vobs + offset to vob x */
    this->p_dvd->i_start = this->p_dvd->i_title_start +
	             DVD_LB_SIZE * (s64)( this->p_dvd->i_sector );

    this->p_dvd->i_chapter = i_chapter;
    return 0;
}

/*****************************************************************************
 *   Find the next sector of a chapter
 *   return 0 ok
 *          -1 : not find (end of the chapter)
 *  Need the beginning of the next chapter,
 *  it seeems to be a reliable way to have the right end.
 *  filepos is :
 *   ( this->p_dvd->i_title_start +
 *                      (s64)( this->p_dvd->i_sector ) *DVD_LB_SIZE );
 *
 *****************************************************************************/

static int DVDFindNextSector(dvd_input_plugin_t *this)
{

    this->p_dvd->i_sector++;

    /* Get the position of the next cell if we're at cell end */

    if( this->p_dvd->i_sector > this->p_dvd->i_end_sector)
    {
        this->p_dvd->i_cell++;

        /* Find cell index in adress map */
        if( DVDFindSector(this) < 0 ) return -1;

        /* End of chapter ? */

         if ( this->p_dvd->i_sector>=this->current_sector_end)
              return -1; /* really the end of the chapter */
         /* FIXME : a bug in the library or a DVD bug? */
         /* Fortunately the cells are in an increase order */

    }

    return 0;
}


 /*****************************************************************************
 *   Select and find the size of a chapter
 *   return : number of  of sector
 *   Need the beginning of the next chapter : current_sector_end
 *   -1 not find
 *
 *****************************************************************************/
 static int DVDGetChapterSize(dvd_input_plugin_t *this, int chapter )

 {
    int size=0;

    if( DVDChapterSelect(this,chapter ) < 0 )
        {
 //           printf ("Can't find chapter\n");
            return -1;
        }

     do {

         size += ( this->p_dvd->i_end_sector - this->p_dvd->i_sector + 1);
         this->p_dvd->i_sector =  this->p_dvd->i_end_sector;

        }
     while (DVDFindNextSector(this)>=0);


      if (DVDChapterSelect(this,chapter)<0 ) return -1;

    return size;
 }

#undef title


static int init_dvd(dvd_input_plugin_t *this) {

     int i,j,k;
     int max_angle;
     int nbchapter;

     if (openDrive (this)==-1) return -1;

     memset(this->current_size,0,sizeof(this->current_size));
     this->current_pos = 0;
     this->current_chapter = 0;
     this->current_title = -1;
     this->current_angle = 1;
     this->current_sector_end=0; /* hack for calculate the real length */
     this->mrl[0]     = 0;
     this->first_play = 1;
     this->branching = 0;
     this->xine_is_seeking=0;

     if( this->p_dvd!=NULL) {
                        if( this->p_dvd->p_css!=NULL) free( this->p_dvd->p_css );
                        if (this->p_dvd->p_ifo!=NULL){
                        if (this->free_ifo) { free(this->p_dvd->p_ifo); }
                         else {
                                IfoDestroy( this->p_dvd->p_ifo );
                                this->free_ifo = 1;
                              }
                        }
                        free( this->p_dvd );
                        this->p_dvd = NULL;
    }

     if( this->p_dvd == NULL ) {
                          this->p_dvd = malloc( sizeof(thread_dvd_data_t) );
                          if (this->p_dvd != NULL) this->p_dvd->p_css  = NULL;
                          if (this->p_dvd != NULL) this->p_dvd->p_ifo = NULL;
    }

    if( this->p_dvd == NULL )
    {
        printf( "dmd error: out of memory\n" );
        return -1;
    }

    this->p_dvd->i_fd = this->dvd_fd;

    i = CSSTest( this->dvd_fd );

    if( i < 0 )
    {
        free(this-> p_dvd );
        this->p_dvd = NULL;
        printf ("Make sure you have a DVD movie in your DVD Player ?\n");
        return -1 ;
    }

    this->p_dvd->b_encrypted = i;

    /* Ifo allocation & initialisation */
    if( IfoCreate( this->p_dvd ) < 0 )
    {
        printf ( "dmd error: allocation error in ifo\n" );
        return -1;
    }

    if( IfoInit(this-> p_dvd->p_ifo ) < 0 )
    {
        printf ( "dmd error: fatal failure in ifo\n" );
        free( this->p_dvd );
        this->p_dvd = NULL;
        return -1;
    }

  /* CSS initialisation */
    if( this->p_dvd->b_encrypted )
    {
        if( this->p_dvd->p_css == NULL )
                       this-> p_dvd->p_css = malloc( sizeof(css_t) );
        if( this->p_dvd->p_css == NULL )
        {
            printf( "dmd error: couldn't create css structure\n" );
            free( this->p_dvd );
            this->p_dvd = NULL;
            return -1;
        }

       this->p_dvd->p_css->i_agid = 0;

        if( CSSInit( this->dvd_fd, this->p_dvd->p_css ) < 0 )
        {
            printf( "dmd error: fatal failure in css\n" );
            free( this->p_dvd->p_css );
            free(this-> p_dvd );
            this->p_dvd = NULL;
            return -1;
        }

        printf ("dmd info: css initialized\n" );
    }

   this->NENTRIES=0;

#define title_inf this->p_dvd->p_ifo->vmg.title_inf
    // printf( "dmd info: number of titles: %d\n", title_inf.i_title_nb );

    /* Built the movie , title which have the max of chapter */

       j=-1;
       k=-1;
       for ( i = 1 ; i <= title_inf.i_title_nb ; i++ ) {
           nbchapter = title_inf.p_attr[i-1].i_chapter_nb;
           if (nbchapter>j) { k=i; j=nbchapter; }
       }

       if (k>0) {
         this->movie.title=k;
         this->movie.start_chapter=1;
         this->movie.end_chapter=j;
         this->movie.angle=1;
         sprintf (this->mrls[this->NENTRIES]->mrl, "dmd://movie");
         this->mrls[this->NENTRIES]->type = ( 0 | mrl_dvd);
         this->mrls[this->NENTRIES]->origin = NULL;
         this->mrls[this->NENTRIES]->link   = NULL;
         this->mrls[this->NENTRIES]->size = 0;
         this->NENTRIES++;
       }

    /* Built the title in the classic way */

    for ( i = 1 ; i <= title_inf.i_title_nb ; i++ )
    {
     nbchapter = title_inf.p_attr[i-1].i_chapter_nb;
     for (j=1; j<= nbchapter;j++) {
                     sprintf (this->mrls[this->NENTRIES]->mrl, "dmd://Title_%d_Chapter_%d",i,j);
                     this->mrls[this->NENTRIES]->type = ( 0 | mrl_dvd);
                     this->mrls[this->NENTRIES]->origin = NULL;
                     this->mrls[this->NENTRIES]->link   = NULL;
                     this->mrls[this->NENTRIES]->size = 0;
                     this->NENTRIES++;
     }
     // printf("Title %d : number of chapter %d\n", i,nbchapter);
    }

   /* Built the Title_x_Chapter_y_Angle_z list z>1 */
   max_angle = 1;
   for ( i = 1 ; i <= title_inf.i_title_nb ; i++ ){
        if (title_inf.p_attr[i-1].i_angle_nb>max_angle)
            max_angle = title_inf.p_attr[i-1].i_angle_nb;
   }

   if (max_angle>1) {
     for (k=2;k<=max_angle;k++){
       for ( i = 1 ; i <= title_inf.i_title_nb ; i++ ){
            nbchapter = title_inf.p_attr[i-1].i_chapter_nb;
            for (j=1; j<= nbchapter;j++) {
              if (k<=title_inf.p_attr[i-1].i_angle_nb) {
                  sprintf (this->mrls[this->NENTRIES]->mrl, "dmd://Title_%d_Chapter_%d_Angle_%d",i,j,k);
                  this->mrls[this->NENTRIES]->type =  ( 0 | mrl_dvd) ;
                  this->mrls[this->NENTRIES]->origin = NULL;
                  this->mrls[this->NENTRIES]->link   = NULL;
                  this->mrls[this->NENTRIES]->size = 0;
                  this->NENTRIES++;
              }
           }
       }
     }
   }
#undef title_inf

  closeDrive(this);

  return 0;
}


static int dvd_plugin_open (input_plugin_t *this_gen, char *mrl) {

  char *filename;
  int d_title;
  int d_chapter;
  int d_angle;
  int m_chapter_size;
  int j;

  dvd_input_plugin_t  *this = (dvd_input_plugin_t *) this_gen;

  /*
   * do we handle this kind of MRL ?
   */

  if (this->xine_is_seeking)
  {
    this->xine_is_seeking=0;
    return 1;
  }

  if (strncasecmp (mrl, "dmd://",6))
    return 0;

  if (this->p_dvd==NULL){
      if (init_dvd(this)!=0) return 0;
  }

  /* Try to find if mrl exist */

  for (j=0;j<this->NENTRIES;j++)
   if (strncasecmp (mrl,this->filelist[j],255)==0) break;

  if (j>=this->NENTRIES) return 0;

  strncpy(this->mrl,this->filelist[j],255);

  filename = &this->mrl[6];

  /* Is movie ? */
  if (strcasecmp (filename, "movie")==0) {

     d_title=this->movie.title;
     d_angle=this->movie.angle;
     d_chapter=this->movie.start_chapter;
     this->is_movie=1;

     printf ("Play the movie Title %d Chapter %d to %d, Angle %d\n",d_title,d_chapter,this->movie.end_chapter,d_angle);

  } else { /* title */

  j = sscanf (filename, "Title_%d_Chapter_%d_Angle_%d", &d_title, &d_chapter,&d_angle);

  if (j==0) return 0;

  if  (j==2) d_angle = 1;

  if (this->branching)
  {
   d_angle=this->current_angle;
   this->branching=0;
  }

  printf ("Play : Title %d Chapter %d Angle %d\n",d_title,d_chapter,d_angle);

  this->is_movie=0;

  }

  if (openDrive(this) < 0) {
    printf ("input_dmd: cannot open dvd drive >%s<\n",this->device);
    return 0;
  }

  /* Drive fd */

  this->p_dvd->i_fd = this->dvd_fd;
  this->p_dvd->p_ifo->i_fd = this->dvd_fd;

  /* Title x Chapter y */

  this->p_dvd->p_ifo->i_title = this->p_dvd->i_title = d_title;
  this->p_dvd->i_chapter = d_chapter;
  this->current_chapter = this->p_dvd->i_chapter;

  /* ifo vts */

if (this->current_title != d_title){

   /* Need to wait ???? */

    usleep(500000);

    if (IfoTitleSet( this->p_dvd->p_ifo ) < 0 )
        {
           this->current_title = -1;
           printf( "dmd error: fatal error in vts ifo\n" );
           closeDrive(this); return 0;
        }

    this->free_ifo = 0;

   /* Need to wait ???? */

   usleep(500000);

 }

  /* title position inside the selected vts */

#define vmg this->p_dvd->p_ifo->vmg
#define vts this->p_dvd->p_ifo->vts

        /* title position inside the selected vts */
   this->p_dvd->i_vts_title =
                    vmg.title_inf.p_attr[this->p_dvd->i_title-1].i_title_num;
   this->p_dvd->i_title_id =
          vts.title_inf.p_title_start[this->p_dvd->i_vts_title-1].i_title_id;
/*
   printf("dvd: title %d vts_title %d pgc %d \n",
                        this->p_dvd->i_title,
                        this->p_dvd->i_vts_title,
                        this->p_dvd->i_title_id );
*/

  /* css title key for current vts */
 if (this->current_title!=d_title){
  if ( this->p_dvd->b_encrypted)
        {
            /* this one is vts number */
            this->p_dvd->p_css->i_title =
                    vmg.title_inf.p_attr[this->p_dvd->i_title-1].i_title_set_num;
            this->p_dvd->p_css->i_title_pos =
                    vts.i_pos +
                    vts.manager_inf.i_title_vob_start_sector * DVD_LB_SIZE;

            j = CSSGetKey(this->p_dvd->i_fd, this->p_dvd->p_css );

            if( j < 0 )
            {
                this->current_title = -1;
                printf( "dmd error: fatal error in vts css key \n" );
                closeDrive(this); return 0;
            }
            else if( j > 0 )
            {
                this->current_title = -1;
                printf( "dmd error: css decryption unavailable\n" );
                closeDrive(this); return 0;
            }

        }
 }

       /*
        * Basic angle management
        */
        this->p_dvd->i_angle = d_angle;

        this->p_dvd->i_angle_nb = vmg.title_inf.p_attr[this->p_dvd->i_title-1].i_angle_nb;

        if( ( this->p_dvd->i_angle <= 0 ) || this->p_dvd->i_angle > this->p_dvd->i_angle_nb )
         {
            this->p_dvd->i_angle = 1;
         }

        this->current_angle = this->p_dvd->i_angle;

         if (this->is_movie==0)
         printf ("dmd info : current angle %d . Number of angle %d\n", this->p_dvd->i_angle,this->p_dvd->i_angle_nb);


        /*
         * Set selected title start and size
         */

        /* title set offset */
        this->p_dvd->i_title_start = vts.i_pos + DVD_LB_SIZE *
                      (s64)( vts.manager_inf.i_title_vob_start_sector );

        /* last video cell */
          this->p_dvd->i_cell = 0;
          this->p_dvd->i_prg_cell = -1 +
          vts.title_unit.p_title[this->p_dvd->i_title_id-1].title.i_cell_nb;

        if( DVDFindCell( this ) < 0 )
        {
            this->current_title = -1;
            printf( "dmd error: can't find title end\n" );
            closeDrive( this ); return 0;
        }

        /* temporary hack to fix size in some dvds */
        if( this->p_dvd->i_cell >= vts.cell_inf.i_cell_nb )
        {
            this->p_dvd->i_cell = vts.cell_inf.i_cell_nb - 1;
        }

        this->p_dvd->i_sector = 0;
        this->p_dvd->i_size = DVD_LB_SIZE *
          (s64)( vts.cell_inf.p_cell_map[this->p_dvd->i_cell].i_end_sector );

//       printf("dmd info: stream size 1: %lld @ %d\n", this->p_dvd->i_size, vts.cell_inf.p_cell_map[this->p_dvd->i_cell].i_end_sector );

      /* FIXME : a bug in the library or a DVD bug?         */
      /* To calculate the size of the chapter */

     if (this->is_movie==1) {

       this->current_sector_end = vts.cell_inf.p_cell_map[this->p_dvd->i_cell].i_end_sector;

     } else {

         if ( DVDChapterSelect( this,d_chapter + 1) < 0 ) {

              this->current_sector_end = vts.cell_inf.p_cell_map[this->p_dvd->i_cell].i_end_sector;

         }  else  { this->current_sector_end = this->p_dvd->i_sector;  }

     }

  /* select and find size of a chapter */

     for (j=1;j<=this->p_dvd->i_angle_nb;j++)
     {
      this->p_dvd->i_angle =j;
      this->current_size[j] = DVDGetChapterSize (this, this->current_chapter) ;
      // printf("Chapter %d Angle %d size %d \n",d_chapter,j,this->current_size[j]);
     }

      this->p_dvd->i_angle = this->current_angle ;
      DVDChapterSelect( this,this->current_chapter);

      m_chapter_size = this->current_size[this->current_angle];

      if(m_chapter_size <0) {
          this->current_title = -1;
          printf( "dmd error: can't set chapter in area\n" );
          closeDrive(this); return 0;
      }

     if (this->is_movie==0) {
        printf("dmd info: chapter %d start at: "FORMATI64"\n",
                                        this->p_dvd->i_chapter, this->p_dvd->i_start);

        printf("dmd info: chapter size "FORMATI64"\n",(s64) m_chapter_size * DVD_LB_SIZE );
     } else {
              printf ("dmd info: movie size "FORMATI64"\n",(s64) m_chapter_size * DVD_LB_SIZE);
            }

     if(m_chapter_size < CHAPTER_SIZE_MIN) {
                        this->current_title = -1;
                        if (this->is_movie==1)
                        printf( "dmd error: movie size too small \n" );
                        else
                        printf( "dmd error: chapter size too small \n" );
                        closeDrive(this); return 0;
      }


     this->current_title = this->p_dvd->i_title;
     this->current_pos = 0;
     this->dsi_valid = 0;
     this->angle_change_request = 0;
     this->angle_change_enabled=0;
     this->angle_available=0;
     this->need_update_title = 1;
  return 1;

}

#undef vts
#undef vmg

/*
 *
 */

static off_t dvd_plugin_seek (input_plugin_t *this_gen, off_t offset, int origin) {
   dvd_input_plugin_t *this = (dvd_input_plugin_t *) this_gen;

   s64 ret;
   int i;

   offset /= DVD_VIDEO_LB_LEN;

  switch (origin) {
      case SEEK_END:
            offset =  this->current_size[this->current_angle] - offset;

      case SEEK_SET:
          if (offset > this->current_size[this->current_angle])  return (off_t) -1;
          this->current_pos = (int) offset;
          DVDChapterSelect(this,this->current_chapter);
          for ( i=0;i<this->current_pos ;i++) DVDFindNextSector(this);
      break;

     case SEEK_CUR:
      if (offset!=0) {
                              this->current_pos += (int) offset;
                              if ( this->current_pos > this->current_size[this->current_angle] ) {
                                  this->current_pos -= (int) offset;
                                  return (off_t) -1;
                              }
                              this->p_dvd->i_sector += (int) offset;
                              if  (this->p_dvd->i_sector > this->p_dvd->i_end_sector) {
                                   DVDChapterSelect(this,this->current_chapter);
                                   for (i=0;i<this->current_pos;i++)  DVDFindNextSector(this);
                               }

      } else {
                  return (off_t ) this->current_pos * (off_t) DVD_VIDEO_LB_LEN;
             }
      break;

      default:
        printf ( "error in input dmd plugin seek:%d is an unknown origin\n",origin);
        return -1;
      break;
  }

  ret = dvd_os_lseek ( this->raw_fd, this->p_dvd->i_title_start +
                       (s64) this->p_dvd->i_sector * DVD_LB_SIZE);

  if ( ret == (s64) -1 ) {
                          printf ("Can't seek file ! \n");
                          return (off_t) -1;
  }

 return ((off_t) this->current_pos * (off_t) DVD_VIDEO_LB_LEN);

}


static off_t dvd_plugin_read (input_plugin_t *this_gen,
			      char *buf, off_t nlen) {
  dvd_input_plugin_t *this = (dvd_input_plugin_t *) this_gen;
  int i,find_angle,offset;

  if (nlen != DVD_VIDEO_LB_LEN) {
    fprintf (stderr, "ERROR in input_dmd plugin read: %lld bytes "
	             "is not a sector!\n", nlen);
    return 0;
  }

  reread:;

  if (this->current_pos>=this->current_size[this->current_angle]) return 0;

  dvd_os_lseek( this->raw_fd,  this->p_dvd->i_title_start + (s64) this->p_dvd->i_sector * (s64) DVD_LB_SIZE);
  if (dvd_os_read (this->raw_fd, buf, DVD_VIDEO_LB_LEN)) {

     if( this->p_dvd->b_encrypted )
        {
          CSSDescrambleSector( this->p_dvd->p_css->pi_title_key,buf);
          (u8) buf[0x14] &= 0x8F;  // ????? a quoi a sert ?????
        }
      if (this->p_dvd->i_angle_nb>1 && is_nav_pack(buf))
        {
          this->dsi_pos=this->p_dvd->i_sector;
          // printf ("NAV PACK at %x\n",this->dsi_pos);
          Parse_DSI(&this->dsi,buf);
          this->dsi_valid = (this->dsi_pos == (int) this->dsi.dsi_gi.nv_pck_lbn);
          // printf("nv_pck_ln = %x\n",this->dsi.dsi_gi.nv_pck_lbn);
          if (this->dsi_valid)
          {
            find_angle=0;

            for (i=0;i<this->p_dvd->i_angle_nb;i++)
            {
              offset = this->dsi.sml_agli.data[i].address;
              if (offset==NO_ILVU_NEXT) { find_angle=2; break; }  /* do nothing */
               else
              if (offset!=0) { find_angle=1; } /* there is angle */
            }
            if (find_angle!=2)
            {
             /* Tell the user  if angle available or not */
              if (this->angle_available!=find_angle)
              {
                this->angle_available=find_angle;
                this->need_update_title=1;
              }
            }
          }
        }

        if (this->dsi_valid && this->angle_change_request)
        {
         offset = this->dsi.sml_agli.data[ this->current_angle - 1].address;
         // printf("offset next_ilvu_start %x\n", offset);
         if (offset!=NO_ILVU_NEXT)
         {
         if (offset==0)
         {
             this->next_ilvu_start = this->p_dvd->i_sector;
         }
         else
         {
             if (offset & 0x80000000)
             this->next_ilvu_start = this->dsi_pos - (offset & 0x7fffffff);
             else
             this->next_ilvu_start = this->dsi_pos + offset;
         }
         } else this->next_ilvu_start = -1;

         offset = this->dsi.sml_agli.data[ this->next_angle_change - 1].address;
         // printf("offset next_ilvu_angle_start %x\n",offset);
         if (offset!=NO_ILVU_NEXT)
         {
         if (offset==0)
         {
             this->next_angle_ilvu_start = this->p_dvd->i_sector;
         }
         else
         {
             if (offset & 0x80000000)
             this->next_angle_ilvu_start = this->dsi_pos - (offset & 0x7fffffff);
             else
             this->next_angle_ilvu_start = this->dsi_pos + offset;
         }
         } else this->next_angle_ilvu_start = -1;

         // printf("Do the change at %x to %x cur pos %x\n", this->next_ilvu_start ,this->next_angle_ilvu_start,this->p_dvd->i_sector);
         this->angle_change_request = 0;

         if (this->next_ilvu_start != -1 && this->next_angle_ilvu_start != -1)
         this->angle_change_enabled = 1;
        }

        if (this->angle_change_enabled && (this->next_ilvu_start==this->p_dvd->i_sector))
        {
          this->next_ilvu_start=-1;
          this->angle_change_enabled = 0;
          this->p_dvd->i_angle=this->current_angle=this->next_angle_change;
          this->current_pos=0;
          DVDChapterSelect(this,this->current_chapter);
          while (this->p_dvd->i_sector<this->next_angle_ilvu_start &&
                 this->current_pos<this->current_size[this->current_angle])
           {
            DVDFindNextSector(this);
            this->current_pos++;
           }
          if (this->current_pos<this->current_size[this->current_angle])
           printf("dmd info: effective switch to angle %d\n",this->current_angle);
          this->need_update_title=1;
          goto reread;
        }
        else
        {
          DVDFindNextSector(this);
          this->current_pos++;
        }

        if (this->need_update_title)
           {
             this->need_update_title=0;
             update_title_display(this);
            }

     return DVD_VIDEO_LB_LEN;

  } else  fprintf (stderr, "input_dmd: read error: %s \n", strerror(errno));

  return 0;
}

/*
 *
 */
static buf_element_t *dvd_plugin_read_block (input_plugin_t *this_gen,
					     fifo_buffer_t *fifo, off_t nlen) {

  dvd_input_plugin_t *this = (dvd_input_plugin_t *) this_gen;
  buf_element_t      *buf;
  s64              pos;
  int i,find_angle,offset;

    if (nlen != DVD_VIDEO_LB_LEN) {
      printf ("input_dmd: error in input_dvd plugin read: %Ld bytes "
      	     "is not a sector!\n", nlen);
    return NULL;
    }

    rereadblock:;
    if (this->current_pos>=this->current_size[this->current_angle]) return NULL;

    pos = this->p_dvd->i_title_start + (s64) this->p_dvd->i_sector * (s64) DVD_LB_SIZE;

    if ((buf = read_cache_read_block (this->read_cache, pos))) {

        buf->type = BUF_DEMUX_BLOCK;

        if( this->p_dvd->b_encrypted )
          {
            CSSDescrambleSector( this->p_dvd->p_css->pi_title_key,buf->content);
            (u8) buf->content[0x14] &= 0x8F;  // ????? a quoi a sert ?????
          }
        if (this->p_dvd->i_angle_nb>1 && is_nav_pack(buf->content))
        {
          this->dsi_pos=this->p_dvd->i_sector;
          // printf ("NAV PACK at %x\n",this->dsi_pos);
          Parse_DSI(&this->dsi,buf->content);
          this->dsi_valid = (this->dsi_pos == (int) this->dsi.dsi_gi.nv_pck_lbn);
          // printf("nv_pck_ln = %x\n",this->dsi.dsi_gi.nv_pck_lbn);

          if (this->dsi_valid)
          {
            find_angle=0;

            for (i=0;i<this->p_dvd->i_angle_nb;i++)
            {
              offset = this->dsi.sml_agli.data[i].address;
              if (offset==NO_ILVU_NEXT) { find_angle=2; break; }  /* do nothing */
               else
              if (offset!=0) { find_angle=1; } /* there is angle */
            }
            if (find_angle!=2)
            {
             /* Tell the user  if angle available or not */
              if (this->angle_available!=find_angle)
              {
                this->angle_available=find_angle;
                this->need_update_title=1;
              }
            }
          }
        }

        if (this->dsi_valid && this->angle_change_request)
        {
         offset = this->dsi.sml_agli.data[this->current_angle - 1].address;
         // printf("offset next_ilvu_start %x\n", offset);
         if (offset!=NO_ILVU_NEXT)
         {
         if (offset==0)
         {
             this->next_ilvu_start = this->p_dvd->i_sector;
         }
         else
         {
             if (offset & 0x80000000)
             this->next_ilvu_start = this->dsi_pos - (offset & 0x7fffffff);
             else
             this->next_ilvu_start = this->dsi_pos + offset;
         }
         } else this->next_ilvu_start = -1;

         offset = this->dsi.sml_agli.data[this->next_angle_change - 1].address;
         // printf("offset next_ilvu_angle_start %x\n",offset);
         if (offset!=NO_ILVU_NEXT)
         {
         if (offset==0)
         {
             this->next_angle_ilvu_start = this->p_dvd->i_sector;
         }
         else
         {
             if (offset & 0x80000000)
             this->next_angle_ilvu_start = this->dsi_pos - (offset & 0x7fffffff);
             else
             this->next_angle_ilvu_start = this->dsi_pos + offset;
         }
         } else this->next_angle_ilvu_start = -1;

         // printf("Do the change at %x to %x cur pos %x\n", this->next_ilvu_start ,this->next_angle_ilvu_start,this->p_dvd->i_sector);
         this->angle_change_request = 0;

         if (this->next_ilvu_start != -1 && this->next_angle_ilvu_start != -1)
            this->angle_change_enabled = 1;
        }

        if (this->angle_change_enabled && (this->next_ilvu_start==this->p_dvd->i_sector))
        {
          this->next_ilvu_start=-1;
          this->angle_change_enabled = 0;
          this->p_dvd->i_angle=this->current_angle=this->next_angle_change;
          this->current_pos=0;
          DVDChapterSelect(this,this->current_chapter);
          while (this->p_dvd->i_sector<this->next_angle_ilvu_start &&
                 this->current_pos<this->current_size[this->current_angle])
           {
            DVDFindNextSector(this);
            this->current_pos++;
           }
          if (this->current_pos<this->current_size[this->current_angle])
          printf("dmd info: switch to angle %d\n",this->current_angle);
          this->need_update_title=1;
          buf->free_buffer(buf);
          goto rereadblock;
        }
        else
        {
         DVDFindNextSector(this);
         this->current_pos++;
        }
    }

   if (this->need_update_title)
   {
      this->need_update_title=0;
      update_title_display(this);
   }
   return buf;

}

/*
 *
 */
static off_t dvd_plugin_get_current_pos (input_plugin_t *this_gen){
  dvd_input_plugin_t *this = (dvd_input_plugin_t *) this_gen;
  return ( (off_t) this->current_pos * (off_t) DVD_VIDEO_LB_LEN);
}

/*
 *
 */
static off_t dvd_plugin_get_length (input_plugin_t *this_gen) {
  dvd_input_plugin_t *this = (dvd_input_plugin_t *) this_gen;
  return ( (off_t) this->current_size[this->current_angle] * (off_t)  DVD_VIDEO_LB_LEN);
}


static uint32_t dvd_plugin_get_capabilities (input_plugin_t *this) {
  return   INPUT_CAP_SEEKABLE  | INPUT_CAP_BLOCK
         | INPUT_CAP_AUTOPLAY  | INPUT_CAP_GET_DIR
         | INPUT_CAP_AUDIOLANG | INPUT_CAP_CLUT;
}

static uint32_t dvd_plugin_get_blocksize (input_plugin_t *this_gen) {
  return DVD_VIDEO_LB_LEN;
}


/*************************************************************************
WIN32 Routine for eject media, load media
*************************************************************************/
#ifdef WIN32

#if !defined (VWIN32_DIOC_DOS_IOCTL)
#define VWIN32_DIOC_DOS_IOCTL      1

typedef struct _DIOC_REGISTERS {
    DWORD reg_EBX;
    DWORD reg_EDX;
    DWORD reg_ECX;
    DWORD reg_EAX;
    DWORD reg_EDI;
    DWORD reg_ESI;
    DWORD reg_Flags;
} DIOC_REGISTERS, *PDIOC_REGISTERS;
#endif

static int eject_load_media_win95(HANDLE fd, BYTE bDrive)
{
   DIOC_REGISTERS regs = {0};
   BOOL  fResult;
   DWORD cb;

   regs.reg_EAX = 0x440D;
   regs.reg_EBX = bDrive;
   regs.reg_ECX = 0x0849;

   fResult = DeviceIoControl (fd, VWIN32_DIOC_DOS_IOCTL,
                              &regs, sizeof(regs), &regs, sizeof(regs),
                              &cb, 0);

   // See if DeviceIoControl and the lock succeeded
   fResult = fResult && !(regs.reg_Flags & 1);

   return fResult;
}

static int eject_load_media_winnt(HANDLE fd)
{
 DWORD ret;
 int res;
 res = DeviceIoControl( fd, IOCTL_STORAGE_EJECT_MEDIA,
                        NULL, 0, NULL, 0, &ret, NULL);
 if (!res)
 {
  res = DeviceIoControl( fd, IOCTL_STORAGE_LOAD_MEDIA,
                        NULL, 0, NULL, 0, &ret, NULL);
 }
 return res;
}
#endif


static int dvd_plugin_eject_media (input_plugin_t *this_gen) {

  dvd_input_plugin_t *this = (dvd_input_plugin_t *) this_gen;
  int ret, status;
  int fd;

#ifdef WIN32
  if (WIN2K)
  {
    char s[16];
    sprintf( s,"\\\\.\\%c:", this->device[0] );
    fd = (int) CreateFile(s,
                           GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE,
                           NULL, OPEN_EXISTING, 0, NULL );
    if((HANDLE) fd == INVALID_HANDLE_VALUE )
    {
      return 0;
    }
    ret = eject_load_media_winnt((HANDLE) fd);
    status = 0;
  }
  else
  {
    fd  = (int) CreateFile ("\\\\.\\vwin32", 0, 0, NULL, 0,FILE_FLAG_DELETE_ON_CLOSE, NULL);
    if((HANDLE) fd == INVALID_HANDLE_VALUE )
    {
      return 0;
    }
    status = (toupper (this->device[0]) - 'A') + 1;
    ret = eject_load_media_win95((HANDLE) fd,status);	
  }
  if (ret)
  {
    printf("input_dmd: open/closing cdrom tray \n");
  }
#else
  if((fd = open(this->device,O_RDONLY | O_NONBLOCK)) != -1) {
#endif

#if defined (HAVE_LINUX_CDROM_H)
    if((status = ioctl(fd, CDROM_DRIVE_STATUS, CDSL_CURRENT)) > 0) {
      switch(status) {
      case CDS_TRAY_OPEN:
        printf("input_dmd: closing cdrom tray\n");
        if((ret = ioctl(fd, CDROMCLOSETRAY)) != 0) {
          fprintf(stderr,"input_dmd: CDROMCLOSETRAY failed: %s\n",strerror(errno));
        }
        break;
      case CDS_NO_INFO:
      case CDS_NO_DISC:
      case CDS_DISC_OK:
        printf("input_dmd: opening cdrom tray\n");
        if((ret = ioctl(fd, CDROMEJECT)) != 0) {
          fprintf(stderr,"input_dmd: CDROMEJECT failed: %s\n",strerror(errno));
        }
        break;
      case CDS_DRIVE_NOT_READY:
        fprintf(stderr, "input_dmd: eject, drive not ready\n");
        break;
      default:
        printf("input_dmd: CDROM_DRIVE_STATUS: unknown drive status: %d\n",status);
        if((ret = ioctl(fd, CDROMEJECT)) != 0) {
          fprintf(stderr,"input_dmd: CDROMEJECT failed: %s\n",strerror(errno));
        }
      }
    }
    else {
      printf("input_dmd: CDROM_DRIVE_STATUS failed: %s\n", strerror(errno));
      close(fd);
      return 0;
    }
#endif

#if defined (HAVE_SYS_CDIO_H)

#ifdef __sun
    status = 0;
    if ((ret = ioctl(fd, CDROMEJECT)) != 0) {
      printf("CDROMEJECT failed: %s\n", strerror(errno));
    }
#else
    if (ioctl(fd, CDIOCALLOW) == -1) {
      perror("ioctl(cdromallow)");
    } else {
      if (ioctl(fd, CDIOCEJECT) == -1) {
        perror("ioctl(cdromeject)");
      }
    }
#endif

#endif

#ifndef WIN32
    close(fd);
 }
#endif
  return 1;
}

 /*
 *
 */
static void dvd_plugin_close (input_plugin_t *this_gen) {
  dvd_input_plugin_t *this = (dvd_input_plugin_t *) this_gen;

  closeDrive (this);

}

/*
 *
 */

static void dvd_plugin_stop (input_plugin_t *this_gen) {

  dvd_input_plugin_t *this = (dvd_input_plugin_t *) this_gen;

  this->xine_is_seeking = 1;

}


/*
 *
 */
static char *dvd_plugin_get_description (input_plugin_t *this_gen) {
  return "DMD css plugin "VERSION;
}

static char *dvd_plugin_get_identifier (input_plugin_t *this_gen) {
  return "DMD";
}

static mrl_t **dvd_plugin_get_dir (input_plugin_t *this_gen,
				   char *filename, int *nEntries){
  dvd_input_plugin_t *this = (dvd_input_plugin_t *) this_gen;

  if (filename) { *nEntries = 0; return NULL; }
  if (init_dvd(this)==-1) { *nEntries = 0; return NULL; }

   *nEntries = this->NENTRIES;
   return this->mrls;
}

static char **dvd_plugin_get_autoplay_list (input_plugin_t *this_gen,
					    int *nFiles) {

    dvd_input_plugin_t *this = (dvd_input_plugin_t *) this_gen;

   if (init_dvd(this)==-1) { *nFiles=0; return NULL; }

   *nFiles = this->NENTRIES;
   return this->filelist;
}


/*
 *
 */
static char * dvd_plugin_get_mrl (input_plugin_t *this_gen) {
  dvd_input_plugin_t *this = (dvd_input_plugin_t *) this_gen;

  return this->mrl;

}


static int dvd_plugin_get_optional_data (input_plugin_t *this_gen,
					 void *data, int data_type) {

  dvd_input_plugin_t *this = (dvd_input_plugin_t *) this_gen;

  int i;
  u8 *dest;
  u32 *src;
  uint16_t lang;
  int channel;

  /* Only if something is open and select */

  if (this->dvd_fd == -1) return  INPUT_OPTIONAL_UNSUPPORTED;

  if (this->current_title == -1) return  INPUT_OPTIONAL_UNSUPPORTED;

  switch(data_type) {

   case INPUT_OPTIONAL_DATA_CLUT:
         // printf("get clut\n");
         if (this->p_dvd == NULL) break;
         if (this->p_dvd->p_ifo == NULL) break;
         if (this->p_dvd->p_ifo->vts.title_unit.p_title==NULL) break;

         dest = (u8 *) data;
         src = (u32 *) this->p_dvd->p_ifo->vts.title_unit.p_title->title.pi_yuv_color;

         /* Put CLUT in the right place */
         for (i=0;i<16;i++) {
                               dest[i*4]=(src[i] >> 24) & 0xff;
                               dest[(i*4)+1]=(src[i] >> 16) & 0xff;
                               dest[(i*4)+2]=(src[i] >> 8 ) & 0xff;
                               dest[(i*4)+3]= src[i] & 0xff;
         }
        /*
        for (i=0;i<64;i++) printf("%02x ",(int) dest[i]);
        printf ("\n");
        */
       return INPUT_OPTIONAL_SUCCESS;
   break;

   case INPUT_OPTIONAL_DATA_AUDIOLANG:
     /* automatic language selection */
         if (this->first_play)
         {
          this->first_play=0;
          if (this->LANG!=0xffff)
          {
            lang=0;
            for (i=1;i<=8;i++)
            {
              lang = IfoGetAudioLang ( this->p_dvd,i);
              if (lang==this->LANG) break;
            }
            /* assume to do a change if default 1<channel<9 */
            if ((i>=2) && (i<=8))
             {
               xine_select_audio_channel(this->xine,i-1);
               sprintf(data,"%c%c", lang >> 8, lang & 0xff);
               printf("DMD set sound channel to (%s) \n",(char *) data);
               return INPUT_OPTIONAL_SUCCESS;
             }
          }
         }

    /* case channel ==-1 or not found : assume default audio channel is 0 */
        channel=0;
        lang = IfoGetAudioLang ( this->p_dvd, channel + 1);
        if(lang != 0xffff) sprintf(data,"%c%c", lang >> 8, lang & 0xff);
        else sprintf (data,"%3d",channel);
        return INPUT_OPTIONAL_SUCCESS;
   break;
  }
  return INPUT_OPTIONAL_UNSUPPORTED;
}

/*
 *
 */

static int dvd_plugin_is_branch_possible (input_plugin_t *this_gen, char *next_mrl) {

  dvd_input_plugin_t *this = (dvd_input_plugin_t *) this_gen;

  char *filename;
  int n_title;
  int n_chapter;
  int n_angle;
  int ret;

  /*
   * do we handle this kind of MRL ?
   */

  if (strncasecmp (next_mrl, "dmd://",6)) {
    printf ("input_dmd: not be a dmd mrl, branching is not possible\n");
    return 0;
  }

  filename = &next_mrl[6];

  ret =  sscanf (filename, "Title_%d_Chapter_%d_Angle_%d", &n_title, &n_chapter,&n_angle);

  if (ret >= 2){
     if ( (n_title==this->p_dvd->i_title) && (n_chapter==(this->p_dvd->i_chapter + 1 )) ) {
        printf ("input_dmd: branching is possible\n");
        this->branching = 1;
        return 1;
     }
  }

  this->branching = 0; /* fail */

  printf ("input_dmd: branching is not possible\n");
  return 0;

}

/*
*
*/
static void event_set_angle (dvd_input_plugin_t *this, int angle)
{
  if (this->p_dvd->i_angle_nb==1) return;
  if (this->p_dvd->i_angle==angle) return;

  if (angle>this->p_dvd->i_angle_nb) angle = 1;
  if (angle<1) angle = this->p_dvd->i_angle_nb;

  this->next_angle_change = angle;
  this->angle_change_request = 1;
  printf("dmd info: User want to switch to angle %d\n", this->next_angle_change);

}

/*
*
*/
static void dmd_event_listener (void *this_gen, xine_event_t *event) {

  dvd_input_plugin_t *this = (dvd_input_plugin_t *) this_gen;

  if(this->dvd_fd==-1) return;

  switch(event->type) {


  case XINE_EVENT_INPUT_ANGLE_NEXT:
       event_set_angle (this,this->current_angle + 1);
  break;

  case XINE_EVENT_INPUT_ANGLE_PREVIOUS:
       event_set_angle (this,this->current_angle - 1);
  break;

  case XINE_EVENT_MOUSE_BUTTON:
  case XINE_EVENT_INPUT_SELECT:
  case XINE_EVENT_INPUT_MENU1:
  break;

  case XINE_EVENT_INPUT_MENU2:
     event_set_angle (this,1);
  break;

  case XINE_EVENT_INPUT_MENU3:
     event_set_angle (this,2);
  break;

  case XINE_EVENT_INPUT_NEXT:
  case XINE_EVENT_INPUT_PREVIOUS:
  case XINE_EVENT_INPUT_UP:
  case XINE_EVENT_INPUT_DOWN:
  case XINE_EVENT_INPUT_LEFT:
  case XINE_EVENT_INPUT_RIGHT:
  case XINE_EVENT_MOUSE_MOVE:
  break;
  }
}

/*
 *
 */

static void dvd_plugin_dispose (input_plugin_t *this_gen ) {
  dvd_input_plugin_t *this = (dvd_input_plugin_t *) this_gen;
  int i;

  if( this->p_dvd!=NULL) {
                        if( this->p_dvd->p_css!=NULL) free( this->p_dvd->p_css );
                        if (this->p_dvd->p_ifo!=NULL){
                        if (this->free_ifo) { free(this->p_dvd->p_ifo); }
                         else {
                                IfoDestroy( this->p_dvd->p_ifo );
                              }
                        }
                        free( this->p_dvd );
                        this->p_dvd = NULL;
  }
  read_cache_free (this->read_cache);

  for (i = 0; i < MAX_DIR_ENTRIES; i++) {
    free ( this->mrls[i]->mrl);
    free ( this->mrls[i]);
  }

  free (this);
}

/*
 *
 */

input_plugin_t *init_input_plugin (int iface, xine_t *xine) {

  dvd_input_plugin_t *this;

  switch (iface) {
  case 8: {

    int i;
    char *s;

    this = (dvd_input_plugin_t *) malloc (sizeof (dvd_input_plugin_t));

    this->dvd_fd =-1;
    this->raw_fd=-1;

    this->p_dvd=NULL;
    this->xine_is_seeking=0;
    this->branching=0;

/* global variables for a chapter */
    this->current_pos = 0;
    this->current_chapter = 0;
    this->current_title = -1;
    this->current_angle = 1;
    this->current_sector_end=0; /* hack for calculate the real length */
    this->free_ifo = 1;
    memset(this->current_size,0,sizeof(this->current_size));

    for (i = 0; i < MAX_DIR_ENTRIES; i++) {

      this->mrls[i]           = (mrl_t *) malloc(sizeof(mrl_t));
      this->mrls[i]->mrl      = (char *) malloc (256);
      this->filelist[i] = this->mrls[i]->mrl;
    }

    this->NENTRIES = 0;

    this->input_plugin.interface_version = INPUT_PLUGIN_IFACE_VERSION;
    this->input_plugin.get_capabilities  = dvd_plugin_get_capabilities;
    this->input_plugin.open              = dvd_plugin_open;
    this->input_plugin.read              = dvd_plugin_read;
    this->input_plugin.read_block        = dvd_plugin_read_block;
    this->input_plugin.seek              = dvd_plugin_seek;
    this->input_plugin.get_current_pos   = dvd_plugin_get_current_pos;
    this->input_plugin.get_length        = dvd_plugin_get_length;
    this->input_plugin.get_blocksize     = dvd_plugin_get_blocksize;
    this->input_plugin.eject_media       = dvd_plugin_eject_media;
    this->input_plugin.close             = dvd_plugin_close;
    this->input_plugin.stop              = dvd_plugin_stop;
    this->input_plugin.get_identifier    = dvd_plugin_get_identifier;
    this->input_plugin.get_description   = dvd_plugin_get_description;
    this->input_plugin.get_dir           = dvd_plugin_get_dir;
    this->input_plugin.get_mrl           = dvd_plugin_get_mrl;
    this->input_plugin.get_autoplay_list = dvd_plugin_get_autoplay_list;
    this->input_plugin.get_optional_data = dvd_plugin_get_optional_data;
    this->input_plugin.is_branch_possible= dvd_plugin_is_branch_possible;
    this->input_plugin.dispose           = dvd_plugin_dispose;

    this->mrl[0]  = 0;
    this->config  = xine->config;
    this->xine    = xine;

    this->first_play=1;

#ifdef WIN32
        s=0;
	this->LANG=LANGIDTOISO639(GetUserDefaultLangID() & 0x3ff);
#else
    if ((s=getenv("LANG"))!=NULL) this->LANG = s[0]<< 8 | s[1];
    else this->LANG=0xffff;
#endif

    // printf ("LANGUAGE=%c%c\n",this->LANG >> 8,this->LANG & 0xff);
    this->device = this->config->register_string(this->config, "dmd.dvd_device",DVD,
                   "DVD device for dmd input plugin",
                    NULL, device_change_cb, (void *)this);
    this->raw_device = this->config->register_string(this->config, "dmd.dvd_raw_device", RDVD,
                    "DVD raw device for dmd input plugin",
                   NULL, device_change_cb, (void *)this);

#ifdef __sun
   check_solaris_vold_device(this);
#endif

    this->read_cache = read_cache_new ();

    xine_register_event_listener(this->xine, dmd_event_listener, this);

{
  cfg_entry_t *entry;
  char *v_mrls;
  entry = this->config->lookup_entry(this->config,"mrl.mrls_mpeg_block");
  if (entry)
  {
   v_mrls = (entry->str_value) ? entry->str_value : entry->unknown_value;
   if(v_mrls)
   {
    if(!strstr(v_mrls, "dmd"))
    {
       if (!entry->str_sticky)
       {
             entry->str_sticky =(char *) malloc(strlen("dmd") + 2);
             sprintf(entry->str_sticky, ",%s", "dmd");
       }
       else
       {
          if(!strstr(entry->str_sticky, "dmd"))
          {
            entry->str_sticky = (char *)
              realloc(entry->str_sticky, strlen(entry->str_sticky) + strlen("dmd") + 2);
              sprintf(entry->str_sticky, "%s,%s", entry->str_sticky,"dmd");
          }
       }
    }
   }
  }
  else
  {

   this->config->register_string(this->config,"mrl.mrls_mpeg_block","dvd,stdin,fifo,dmd",
					"valid mrls for mpeg block demuxer",
					NULL, NULL, NULL);
  }
}

    printf ("DMD css plugin %s init\n",VERSION);

    return (input_plugin_t *) this;
  }
    break;

  default:
    fprintf(stderr,
	    "This DVD input plugin (DMD)  doesn't support plugin API version %d.\n"
	    "PLUGIN DISABLED.\n"
	    "This means there's a version mismatch between xine and this input"
	    "plugin.\nInstalling current input plugins should help.\n",
	    iface);
    return NULL;
  }
}



