/**
 * Copyright (C) 2000 Hkan Hjort <d95hjort@dtek.chalmers.se>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include <inttypes.h>

#include "parse_dsi.h"

uint8_t ReadByte (unsigned char **pp_current)
 {
   uint8_t i_ret;
   i_ret = *(*pp_current);
   (*pp_current)++;
   return i_ret;
 }

uint16_t ReadWord (unsigned char **pp_current)
 {
   uint16_t i_ret;
   i_ret = ReadByte (pp_current);
   i_ret <<=8;
   i_ret |= ReadByte (pp_current);
   return i_ret;
 }

uint32_t ReadDouble (unsigned char **pp_current)
 {
   uint32_t i_ret;
   i_ret = ReadByte (pp_current);
   i_ret <<=8;
   i_ret |= ReadByte (pp_current);
   i_ret <<=8;
   i_ret |= ReadByte (pp_current);
   i_ret <<=8;
   i_ret |= ReadByte (pp_current);
   return i_ret;
 }

int  Parse_DSI(dsi_t *dsi, unsigned char *buffer) {
  int i;
  unsigned char *pp_current=&buffer[DSI_START_BYTE];

  /* dsi dsi gi */
  dsi->dsi_gi.nv_pck_scr = ReadDouble (&pp_current);
  dsi->dsi_gi.nv_pck_lbn = ReadDouble (&pp_current);
  dsi->dsi_gi.vobu_ea = ReadDouble (&pp_current);
  dsi->dsi_gi.vobu_1stref_ea = ReadDouble (&pp_current);
  dsi->dsi_gi.vobu_2ndref_ea = ReadDouble (&pp_current);
  dsi->dsi_gi.vobu_3rdref_ea = ReadDouble (&pp_current);
  dsi->dsi_gi.vobu_vob_idn = ReadWord (&pp_current);
  dsi->dsi_gi.zero1= ReadByte (&pp_current);
  dsi->dsi_gi.vobu_c_idn = ReadByte (&pp_current);
  dsi->dsi_gi.c_eltm.hour = ReadByte (&pp_current);
  dsi->dsi_gi.c_eltm.minute = ReadByte (&pp_current);
  dsi->dsi_gi.c_eltm.second = ReadByte (&pp_current);
  dsi->dsi_gi.c_eltm.frame_u = ReadByte (&pp_current);

  /* dsi sml pbi */
  dsi->sml_pbi.category = ReadWord (&pp_current);
  dsi->sml_pbi.ilvu_ea = ReadDouble (&pp_current);
  dsi->sml_pbi.ilvu_sa = ReadDouble (&pp_current);
  dsi->sml_pbi.size = ReadWord (&pp_current);
  dsi->sml_pbi.vob_v_s_s_ptm = ReadDouble (&pp_current);
  dsi->sml_pbi.vob_v_e_e_ptm = ReadDouble (&pp_current);
  for(i = 0; i < 8; i++){
    dsi->sml_pbi.vob_a[i].stp_ptm1 = ReadDouble (&pp_current);
    dsi->sml_pbi.vob_a[i].stp_ptm2 = ReadDouble (&pp_current);
    dsi->sml_pbi.vob_a[i].gap_len1 = ReadDouble (&pp_current);
    dsi->sml_pbi.vob_a[i].gap_len2 = ReadDouble (&pp_current);
  }

  /* dsi sml agli */
  for(i = 0; i < 9; i++) {
     dsi->sml_agli.data[ i ].address = ReadDouble (&pp_current);
     dsi->sml_agli.data[ i ].size = ReadWord (&pp_current);
  }

  /* dsi vobu sri */
  dsi->vobu_sri.next_video = ReadDouble (&pp_current);
  for(i = 0; i < 19; i++) dsi->vobu_sri.fwda[i] = ReadDouble (&pp_current);
  dsi->vobu_sri.next_vobu = ReadDouble (&pp_current);
  dsi->vobu_sri.prev_vobu = ReadDouble (&pp_current);
  for(i = 0; i < 19; i++) dsi->vobu_sri.bwda[i] = ReadDouble (&pp_current);
  dsi->vobu_sri.prev_video = ReadDouble (&pp_current);

  /* dsi synci */
  for(i = 0; i < 8; i++) dsi->synci.a_synca[i] = ReadWord (&pp_current);
  for(i = 0; i < 32; i++) dsi->synci.sp_synca[i] = ReadDouble (&pp_current);

  /* zero nothing */

  /* dsi dsi gi */
  
  if (dsi->dsi_gi.zero1 != 0)
   {
    return -1;
    // printf ("Parse dsi : This packet seems bad \n");
    // Not use 
   }

   return 0;
}

/**
 * Returns true if the pack is a NAV pack.  This check is clearly insufficient,
 * and sometimes we incorrectly think that valid other packs are NAV packs.  I
 * need to make this stronger.
 */

int is_nav_pack( unsigned char *buffer )
{
  /* I say just have a pci and dsi packet */ 
    return ( buffer[  41  ] == 0xbf &&
             buffer[  44  ] == 0x00 &&
             buffer[ 1027 ] == 0xbf &&
             buffer[ 1030 ] == 0x01
             );
}





